/// <reference types="angular" />
/** @module ng2_directives */ /** */
import { ViewContainerRef, ComponentRef, Injector } from '@angular/core';
import { UIRouter } from "../../router";
import { ViewContext, ViewConfig, ActiveUIView } from "../../view/interface";
import { Ng2ViewConfig } from "../statebuilders/views";
import { ResolveContext } from "../../resolve/resolveContext";
export interface ParentUIViewInject {
    context: ViewContext;
    fqn: string;
}
/**
 * A UI-Router viewport directive, which is filled in by a view (component) on a state.
 *
 * ### Selector
 *
 * A `ui-view` directive can be created as an element: `<ui-view></ui-view>` or as an attribute: `<div ui-view></div>`.
 *
 * ### Purpose
 *
 * This directive is used in a Component template (or as the root component) to create a viewport.  The viewport
 * is filled in by a view (as defined by a [[Ng2ViewDeclaration]] inside a [[Ng2StateDeclaration]]) when the view's
 * state has been activated.
 *
 * @example
 * ```js
 *
 * // This app has two states, 'foo' and 'bar'
 * stateRegistry.register({ name: 'foo', url: '/foo', component: FooComponent });
 * stateRegistry.register({ name: 'bar', url: '/bar', component: BarComponent });
 * ```
 * ```html
 * <!-- This ui-view will be filled in by the foo state's component or
 *      the bar state's component when the foo or bar state is activated -->
 * <ui-view></ui-view>
 * ```
 *
 * ### Named ui-views
 *
 * A `ui-view` may optionally be given a name via the attribute value: `<div ui-view='header'></div>`.  *Note:
 * an unnamed `ui-view` is internally named `$default`*.   When a `ui-view` has a name, it will be filled in
 * by a matching named view.
 *
 * @example
 * ```js
 *
 * stateRegistry.register({
 *   name: 'foo',
 *   url: '/foo',
 *   views: { header: HeaderComponent, $default: FooComponent });
 * ```
 * ```html
 * <!-- When 'foo' state is active, filled by HeaderComponent -->
 * <div ui-view="header"></div>
 *
 * <!-- When 'foo' state is active, filled by FooComponent -->
 * <ui-view></ui-view>
 * ```
 */
export declare class UIView {
    router: UIRouter;
    viewContainerRef: ViewContainerRef;
    componentTarget: ViewContainerRef;
    name: string;
    _name: string;
    componentRef: ComponentRef<any>;
    deregister: Function;
    uiViewData: ActiveUIView;
    parent: ParentUIViewInject;
    static PARENT_INJECT: string;
    constructor(router: UIRouter, parent: any, viewContainerRef: ViewContainerRef);
    ngOnInit(): void;
    disposeLast(): void;
    ngOnDestroy(): void;
    /**
     * The view service is informing us of an updated ViewConfig
     * (usually because a transition activated some state and its views)
     */
    viewConfigUpdated(config: ViewConfig): void;
    applyUpdatedConfig(config: Ng2ViewConfig): void;
    /**
     * Creates a new Injector for a routed component.
     *
     * Adds resolve values to the Injector
     * Adds providers from the NgModule for the state
     * Adds providers from the parent Component in the component tree
     * Adds a PARENT_INJECT view context object
     *
     * @returns an Injector
     */
    getComponentInjector(context: ResolveContext): Injector;
    /**
     * Supplies component inputs with resolve data
     *
     * Finds component inputs which match resolves (by name) and sets the input value
     * to the resolve data.
     */
    applyInputBindings(ref: ComponentRef<any>, context: ResolveContext, componentClass: any): void;
}
